from cuml import preprocessing
from sklearn.preprocessing import StandardScaler

from optimus.engines.base.commons.functions import string_to_index, index_to_string, find, to_float_cudf, \
    to_string_cudf, to_integer_cudf
from optimus.engines.base.cudf.columns import CUDFBaseColumns
from optimus.engines.base.dataframe.columns import DataFrameBaseColumns
from optimus.engines.base.columns import BaseColumns
from optimus.engines.base.meta import Meta
from optimus.helpers.columns import parse_columns, get_output_cols
from optimus.helpers.constants import Actions
from optimus.helpers.core import val_to_list
from optimus.helpers.raiseit import RaiseIt
from optimus.infer import is_list_of_tuples


class Cols(CUDFBaseColumns, DataFrameBaseColumns, BaseColumns):

    def __init__(self, df):
        super().__init__(df)

    @property
    def _pd(self):
        import cudf
        return cudf

    def _series_to_pandas(self, series):
        return series.to_pandas()

    @staticmethod
    def exec_agg(exprs, compute):
        """
        Execute and aggregation
        :param exprs:
        :return:
        """
        # print("exprs",type(exprs))
        try:
            return exprs[0].to_pandas().to_dict()
        except TypeError:
            return exprs
        except AttributeError:
            return exprs
        except IndexError:
            return exprs

    def find(self, columns, sub, ignore_case=False):
        """
        Find the start and end position for a char or substring
        :param columns:
        :param ignore_case:
        :param sub:
        :return:
        """
        import cudf

        df = self.root
        # TODO: This could be slow try to implement the find function in cudf
        df = df.to_pandas()

        df = find(df, columns, sub, ignore_case)

        return cudf.from_pandas(df)

    def to_string(self, input_cols="*", output_cols=None):
        df = self.root
        return df.cols.apply(input_cols, to_string_cudf, output_cols=output_cols,
                             meta_action=Actions.TO_STRING.value,
                             mode="vectorized")

    def to_integer(self, input_cols="*", output_cols=None):

        df = self.root
        return df.cols.apply(input_cols, to_integer_cudf, output_cols=output_cols,
                             meta_action=Actions.TO_INTEGER.value,
                             mode="pandas")

    @staticmethod
    def to_timestamp(input_cols, date_format=None, output_cols=None):
        pass

    def reverse(self, input_cols, output_cols=None):
        raise NotImplementedError('Not implemented yet')

    # https://github.com/rapidsai/cudf/issues/3177
    # def replace(self, input_cols, search=None, replace_by=None, search_by="chars", output_cols=None):
    #     """
    #     Replace a value, list of values by a specified string
    #     :param input_cols: '*', list of columns names or a single column name.
    #     :param output_cols:
    #     :param search: Values to look at to be replaced
    #     :param replace_by: New value to replace the old one
    #     :param search_by: Can be "full","words","chars" or "numeric".
    #     :return: Dask DataFrame
    #     """
    #
    #     df = self.df
    #     columns = prepare_columns(df, input_cols, output_cols)
    #
    #     search = val_to_list(search)
    #     if search_by == "chars":
    #         _regex = search
    #         # _regex = re.compile("|".join(map(re.escape, search)))
    #     elif search_by == "words":
    #         _regex = (r'\b%s\b' % r'\b|\b'.join(map(re.escape, search)))
    #     else:
    #         _regex = search
    #
    #     if not is_str(replace_by):
    #         RaiseIt.type_error(replace_by, ["str"])
    #
    #     kw_columns = {}
    #     for input_col, output_col in columns:
    #         if search_by == "chars" or search_by == "words":
    #             # This is only implemented in cudf
    #             kw_columns[output_col] = df[input_col].astype(str).str.replace_multi(search, replace_by,
    #                                                                                  regex=False)
    #         elif search_by == "full":
    #             kw_columns[output_col] = df[input_col].astype(str).replace(search, replace_by)
    #
    #     print("kw_columns",kw_columns)
    #     print("search",search)
    #     df.assign(**kw_columns)
    #     return df

    def replace_regex(self, input_cols, regex=None, replace_by=None, output_cols=None):
        df = self.root
        input_cols = parse_columns(df, input_cols)
        output_cols = get_output_cols(input_cols, output_cols)
        for input_col, output_col in zip(input_cols, output_cols):
            df[output_col] = df[input_col].str.replace_multi(regex, replace_by, regex=True)
        return df

    # NLP
    def stem_words(self, input_col):
        raise NotImplementedError('Not implemented yet')


    def strip_html(self):
        df = self.root
        # soup = BeautifulSoup(self.text, "html.parser")
        # self.text = soup.get_text()
        raise NotImplementedError('Not implemented yet')

    def min_max_scaler(self, input_cols, output_cols=None):
        pass

    def standard_scaler(self, input_cols, output_cols=None):

        df = self.root
        scaler = StandardScaler()
        for input_col, output_col in zip(input_cols, output_cols):
            data = df[input_col]
            scaler.fit(data)
            df[output_col] = scaler.transform(data)

        return df

    def impute(self, input_cols, data_type="continuous", strategy="mean", fill_value=None, output_cols=None):
        """

        :param input_cols:
        :param data_type:
        :param strategy:
        # - If "mean", then replace missing values using the mean along
        #   each column. Can only be used with numeric data.
        # - If "median", then replace missing values using the median along
        #   each column. Can only be used with numeric data.
        # - If "most_frequent", then replace missing using the most frequent
        #   value along each column. Can be used with strings or numeric data.
        # - If "constant", then replace missing values with fill_value. Can be
        #   used with strings or numeric data.
        :param output_cols:
        :return:
        """

        raise NotImplementedError("Not implemented yet")

    @staticmethod
    def max_abs_scaler(input_cols, output_cols=None):
        raise NotImplementedError('Not implemented yet')

    def is_match(self, columns, dtype, invert=False):
        """
        Find the rows that match a data type
        :param columns:
        :param dtype: data type to match
        :param invert: Invert the match
        :return:
        """
        df = self.root
        columns = parse_columns(df, columns)

        f = profiler_dtype_func(dtype)
        if f is not None:
            for col_name in columns:
                df = df[col_name].to_pandas().apply(f)
                df = ~df if invert is True else df
            return df

    @staticmethod
    def heatmap(columns, buckets=10):
        pass

    def hist(self, columns="*", buckets=20, compute=True):
        import cupy as cp
        df = self.root
        columns = parse_columns(df, columns)

        result = {}
        for col_name in columns:
            # Seems to be some kind of bug of some types of data sets that converts all values to nan,
            # we drop na before  converting to array
            df_numeric = cp.array(to_float_cudf(df.data[col_name]).dropna().to_gpu_array())
            if len(df_numeric) > 0:
                _count, _bins = cp.histogram(df_numeric, buckets)
                result[col_name] = [
                    {"lower": float(_bins[i]), "upper": float(_bins[i + 1]), "count": int(_count[i])}
                    for i in range(buckets)]

        return {"hist": result}

    def count_by_dtypes(self, columns, infer=False, str_funcs=None, int_funcs=None):
        df = self.root
        result = {}
        df_len = len(df)
        for col_name, na in df.cols.count_na(columns).items():
            result[col_name] = {"match": df_len - na, "missing": na, "mismatches": 0}
        return result

        # return np.count_nonzero(df.isnull().values.ravel())

    def qcut(self, columns, quantiles, handle_invalid="skip"):
        pass

    def string_to_index(self, input_cols=None, output_cols=None, columns=None):
        df = self.root
        le = preprocessing.LabelEncoder()
        return string_to_index(df, input_cols, output_cols, le)

    def index_to_string(self, input_cols=None, output_cols=None, columns=None):
        df = self.root
        le = preprocessing.LabelEncoder()
        return index_to_string(df, input_cols, output_cols, le)

    def _unnest(self, dfd, input_col, final_columns, separator, splits, mode, output_cols):
        if mode == "string":
            dfd_new = dfd[input_col].astype(str).str.split(separator, expand=True, n=splits)

        else:
            RaiseIt.value_error(mode, ["string"])

        return dfd_new



# -*- coding: utf-8 -*-
# ---
# jupyter:
#   jupytext:
#     formats: ipynb,py:light
#     text_representation:
#       extension: .py
#       format_name: light
#       format_version: '1.4'
#       jupytext_version: 1.1.1
#   kernel_info:
#     name: python3
#   kernelspec:
#     display_name: Python 3
#     language: python
#     name: python3
# ---

# # This notebook create the tests in python code. All this cells must be run to executed the tests

# %load_ext autoreload
# %autoreload 2

# + {"outputHidden": false, "inputHidden": false}
import sys
sys.path.append("../..")
# -

from optimus import Optimus
from optimus.helpers.test import Test

from optimus import Optimus
op = Optimus("dask", n_workers=1, threads_per_worker=8, processes=False, memory_limit="3G", comm=True)

import numpy as np
import pandas as pd


# +
import pandas as pd
from pyspark.sql.types import *
from datetime import date, datetime

cols = [
    ("names", "str"),
    ("height(ft)", ShortType()),
    ("function", "str"),
    ("rank", ByteType()),
    ("age", "int"),
    ("weight(t)", "float"),
    "japanese name",
    "last position seen",
    "date arrival",
    "last date seen",
    ("attributes", ArrayType(FloatType())),
    ("Date Type", DateType()),
    ("timestamp", TimestampType()),
    ("Cybertronian", BooleanType()),
    ("function(binary)", BinaryType()),
    ("NullType", NullType())

]

rows = [
    ("Optim'us", -28, "Leader", 10, 4000000, 4.30, ["Inochi", "Convoy"], "19.442735,-99.201111", "1980/04/10",
     "2016/09/10", [8.5344, 4300.0], date(2016, 9, 10), datetime(2014, 6, 24), True, bytearray("Leader", "utf-8"),
     None),
    ("bumbl#ebéé  ", 17, "Espionage", 7, 5000000, 2.0, ["Bumble", "Goldback"], "10.642707,-71.612534", "1980/04/10",
     "2015/08/10", [5.334, 2000.0], date(2015, 8, 10), datetime(2014, 6, 24), True, bytearray("Espionage", "utf-8"),
     None),
    ("ironhide&", 26, "Security", 7, 7000000, 4.0, ["Roadbuster"], "37.789563,-122.400356", "1980/04/10",
     "2014/07/10", [7.9248, 4000.0], date(2014, 6, 24), datetime(2014, 6, 24), True, bytearray("Security", "utf-8"),
     None),
    ("Jazz", 13, "First Lieutenant", 8, 5000000, 1.80, ["Meister"], "33.670666,-117.841553", "1980/04/10",
     "2013/06/10", [3.9624, 1800.0], date(2013, 6, 24), datetime(2014, 6, 24), True,
     bytearray("First Lieutenant", "utf-8"), None),
    ("Megatron", None, "None", 10, 5000000, 5.70, ["Megatron"], None, "1980/04/10", "2012/05/10", [None, 5700.0],
     date(2012, 5, 10), datetime(2014, 6, 24), True, bytearray("None", "utf-8"), None),
    ("Metroplex_)^$", 300, "Battle Station", 8, 5000000, None, ["Metroflex"], None, "1980/04/10", "2011/04/10",
     [91.44, None], date(2011, 4, 10), datetime(2014, 6, 24), True, bytearray("Battle Station", "utf-8"), None),
    (None, None, None, np.nan, None, None, None, None, None, None, None, None, None, None, None, None),

]

source_df = pd.DataFrame(columns = ['names','height(ft)','function',
                      'rank','age','weight(t)','japanese name','last position seen',
                      'date arrival','last date seen','attributes','Date Type','timestamp',
                      'Cybertronian','function(binary)','NullType'], 
             data= rows,
#              dtype = (object,object,object,
#                                 object,object,object,object,object,
#                                 object,object,object,object,object,
#                                 object,object)
            )

# from dask import dataframe as dd

# a=pd.DataFrame(source_df.to_dict())
from dask import dataframe as dd
source_df = dd.from_pandas(source_df, npartitions=1)


# -

source_df.compute()

# ### End Init Section

# # Test

# ## Columns Test

t = Test(op, source_df, "df_cols_dask", imports=["import numpy as np",
                                            "nan = np.nan",
                                            "import datetime",], path="df_cols_dask", final_path="..")

# +
from pyspark.sql import functions as F


def func(col_name, attrs):
    return F.col(col_name) * 2

numeric_col = "height(ft)"
numeric_col_B = "rank"
numeric_col_C = "rank"
string_col = "function"
date_col = "date arrival"
date_col_B = "last date seen"
new_col = "new col"
array_col = "attributes"
# -

source_df.compute()

t.create(None, "cols.clip", "all_columns", "df", None,"*", 3, 5)

t.create(None, "cols.cast", "all_columns", "df", None, "*", "str")

t.create(None, "cols.is_na", "all_columns", "df", None, "*")

for i in source_df.compute().attributes:
    print(type(i[0]))

source_df.cols.unnest("attributes").compute()

source_df.compute()

source_df['height(ft)'].astype(str).str.split(".").compute()

import dask
source_df['height(ft)'].astype(str).str.split(".", expand=True, n=1)
# apply(dask.dataframe.Series, meta=("a")).compute()

source_df['attributes'].compute().apply(pd.Series)

source_df = source_df.repartition(6)

source_df["date arrival"].astype(str).str.split("/", expand=True, n=2).compute()

source_df.compute()



from enum import Enum

from optimus.helpers.logger import logger
from optimus.helpers.raiseit import RaiseIt
import nltk

# if importlib.util.find_spec("vaex") is not None:
#     from vaex import DataFrame as VaexDataFrame
#     # import pandas as pd
#     from optimus.engines.vaex import rows, columns, extension, constants, functions
#     from optimus.engines.vaex.io import save
#
#     VaexDataFrame.outliers = property(outliers)
#     VaexDataFrame.meta = property(meta)
#     VaexDataFrame.schema = [MetadataDask()]


class Engine(Enum):
    PANDAS = "pandas"
    CUDF = "cudf"
    DASK = "dask"
    DASK_CUDF = "dask_cudf"
    SPARK = "spark"
    VAEX = "vaex"
    IBIS = "ibis"

    @classmethod
    def list(cls):
        return list(map(lambda c: c.value, cls))


def optimus(engine=Engine.DASK.value, *args, **kwargs):
    """
    This is the entry point to initialize the selected engine.
    :param engine: A string identifying an engine :classL`Engine`.
    :param args:
    :param kwargs:
    :return:
    """
    logger.print("ENGINE", engine)

    # lemmatizer
    nltk.download('wordnet', quiet= True)

    # Stopwords
    nltk.download('stopwords', quiet=True)

    # POS
    nltk.download('averaged_perceptron_tagger')

    # Init engine
    if engine == Engine.PANDAS.value:
        from optimus.engines.pandas.engine import PandasEngine
        op = PandasEngine(*args, **kwargs)

    elif engine == Engine.VAEX.value:
        from optimus.engines.vaex.engine import VaexEngine
        op = VaexEngine(*args, **kwargs)

    elif engine == Engine.SPARK.value:
        from optimus.engines.spark.engine import SparkEngine
        op = SparkEngine(*args, **kwargs)

    elif engine == Engine.DASK.value:
        from optimus.engines.dask.engine import DaskEngine
        op = DaskEngine(*args, **kwargs)

    elif engine == Engine.IBIS.value:
        from optimus.engines.ibis.engine import IbisEngine
        op = IbisEngine(*args, **kwargs)

    elif engine == Engine.CUDF.value:
        from optimus.engines.cudf.engine import CUDFEngine
        op = CUDFEngine(*args, **kwargs)

    elif engine == Engine.DASK_CUDF.value:
        from optimus.engines.dask_cudf.engine import DaskCUDFEngine
        op = DaskCUDFEngine(*args, **kwargs)

    else:
        RaiseIt.value_error(engine, Engine.list())

    # Set cupy yo user RMM
    def switch_to_rmm_allocator():
        import rmm
        import cupy
        cupy.cuda.set_allocator(rmm.rmm_cupy_allocator)
        return True

    if engine == Engine.CUDF.value:
        switch_to_rmm_allocator()

    if engine == Engine.DASK_CUDF.value:
        if op.client:
            op.client.run(switch_to_rmm_allocator)

    return op



import os

from optimus.helpers.logger import logger
from optimus.helpers.functions import prepare_path_local, path_is_local


class Save:
    def __init__(self, root):
        self.root = root

    def json(self, path, storage_options=None, conn=None, *args, **kwargs):
        """
        Save data frame in a json file
        :param path: path where the spark will be saved.
        :param mode: Specifies the behavior of the save operation when data already exists.
                "append": Append contents of this DataFrame to existing data.
                "overwrite" (default case): Overwrite existing data.
                "ignore": Silently ignore this operation if data already exists.
                "error": Throw an exception if data already exists.
        :param num_partitions: the number of partitions of the DataFrame
        :return:
        """
        
        df = self.root.data

        if conn is not None:
            path = conn.path(path)
            storage_options = conn.storage_options

        try:
            os.makedirs(path, exist_ok=True)
            df.to_json(filename=path, storage_options=storage_options, *args, **kwargs)
        except (OSError, IOError) as error:
            logger.print(error)
            raise
        #     print("Creation of the directory %s failed" % path)
        # else:
        #     print("Successfully created the directory %s" % path)

    def csv(self, path, mode="wt", index=False, single_file=True, storage_options=None, conn=None, **kwargs):
        """
        Save data frame to a CSV file.
        :param mode: 'rb', 'wt', etc
        :param single_file:
        :param index:
        :param path: path where the spark will be saved.
        it uses the default value.
        :return: Dataframe in a CSV format in the specified path.
        """

        df = self.root.data

        if conn is not None:
            path = conn.path(path)
            storage_options = conn.storage_options

        try:
            if path_is_local(path):
                prepare_path_local(path)
            
            df.to_csv(filename=path, mode=mode, index=index, single_file=single_file, storage_options=storage_options, **kwargs)
        except IOError as error:
            logger.print(error)
            raise

    def parquet(self, path, mode="overwrite", num_partitions=1, engine="pyarrow", storage_options=None, conn=None, **kwargs):
        """
        Save data frame to a parquet file
        :param path: path where the spark will be saved.
        :param mode: Specifies the behavior of the save operation when data already exists.
                    "append": Append contents of this DataFrame to existing data.
                    "overwrite" (default case): Overwrite existing data.
                    "ignore": Silently ignore this operation if data already exists.
                    "error": Throw an exception if data already exists.
        :param num_partitions: the number of partitions of the DataFrame
        :return:
        """

        # This character are invalid as column names by parquet
        invalid_character = [" ", ",", ";", "{", "}", "(", ")", "\n", "\t", "="]

        def func(col_name):
            for i in invalid_character:
                col_name = col_name.replace(i, "_")
            return col_name

        df = self.root.cols.rename(func)

        if conn is not None:
            path = conn.path(path)
            storage_options = conn.storage_options
        dfd = df.data
        try:
            if engine == 'pyarrow':
                dfd.to_parquet(path,  engine='pyarrow', storage_options=storage_options, **kwargs)
            elif engine == "fastparquet":
                dfd.to_parquet(path, engine='fastparquet', storage_options=storage_options, **kwargs)

        except IOError as e:
            logger.print(e)
            raise

    @staticmethod
    def avro(path):
        raise NotImplementedError('Not implemented yet')



from optimus.engines.base.io.drivers.impala import ImpalaDriver
from optimus.helpers.raiseit import RaiseIt
from optimus.engines.base.io.drivers.abstract_driver import AbstractDriver
from optimus.engines.base.io.drivers.cassandra import CassandraDriver
from optimus.engines.base.io.drivers.mysql import MySQLDriver
from optimus.engines.base.io.drivers.oracle import OracleDriver
from optimus.engines.base.io.drivers.postgresql import PostgreSQLDriver
from optimus.engines.base.io.drivers.presto import PrestoDriver
from optimus.engines.spark.io.properties import DriverProperties
from optimus.engines.base.io.drivers.redshift import RedshiftDriver
from optimus.engines.base.io.drivers.sqlite import SQLiteDriver
from optimus.engines.base.io.drivers.sqlserver import SQLServerDriver
from optimus.engines.base.io.drivers.bigquery import BigQueryDriver


class DriverFactory:
    """Database driver factory. This database driver factory currently supports the following implementations:

        * Cassandra
        * MySQL
        * Oracle
        * Postgres
        * Presto
        * Redshift
        * SQLite
        * SQLServer
        * BigQuery
    """

    @staticmethod
    def get(driver_type) -> AbstractDriver:
        """
        Returns a driver implementation given a database name

        :param driver_type: name of the database
        :return: a database driver
        """
        if driver_type == DriverProperties.CASSANDRA.value["name"]:
            return CassandraDriver()
        elif driver_type == DriverProperties.MYSQL.value["name"]:
            return MySQLDriver()
        elif driver_type == DriverProperties.ORACLE.value["name"]:
            return OracleDriver()
        elif driver_type == DriverProperties.POSTGRESQL.value["name"]:
            return PostgreSQLDriver()
        elif driver_type == DriverProperties.PRESTO.value["name"]:
            return PrestoDriver()
        elif driver_type == DriverProperties.REDSHIFT.value["name"]:
            return RedshiftDriver()
        elif driver_type == DriverProperties.SQLITE.value["name"]:
            return SQLiteDriver()
        elif driver_type == DriverProperties.SQLSERVER.value["name"]:
            return SQLServerDriver()
        elif driver_type == DriverProperties.BIGQUERY.value["name"]:
            return BigQueryDriver()
        elif driver_type == DriverProperties.IMPALA.value["name"]:
            return ImpalaDriver()
        else:
            RaiseIt.value_error(driver_type, [database["name"] for database in DriverProperties.list()])



# Reference https://stackoverflow.com/questions/46692370/usage-of-custom-python-object-in-pyspark-udf
# def parse_to_profiler_dtypes(df, col_name):
class test_pickle(object):
    def __call__(self, *args, **kwargs):
        print(args, kwargs)




import ntpath

import dask.bag as dask_bag
import pandas as pd
from dask import dataframe as dd

import optimus.helpers.functions_spark
from optimus.engines.base.io.load import BaseLoad
from optimus.engines.base.meta import Meta
from optimus.engines.dask.dataframe import DaskDataFrame
from optimus.helpers.core import val_to_list
from optimus.helpers.functions import prepare_path, unquote_path
from optimus.helpers.logger import logger


class Load(BaseLoad):

    def __init__(self, op):
        self.op = op

    class test:
        def __init__(self):
            pass

    @staticmethod
    def json(path, multiline=False, lines=True, conn=None, storage_options=None, *args, **kwargs):
        """
        Return a dask dataframe from a json file.
        :param path: path or location of the file.
        :param multiline:

        :return:
        """

        path = unquote_path(path)

        if conn is not None:
            path = conn.path(path)
            storage_options = conn.storage_options

        try:
            # TODO: Check a better way to handle this Spark.instance.spark. Very verbose.
            dfd = dd.read_json(path, lines=lines, storage_options=storage_options, *args, **kwargs)
            df = DaskDataFrame(dfd)
            df.meta = Meta.set(df.meta, value={"file_name": path, "name": ntpath.basename(path)})

        except IOError as error:
            logger.print(error)
            raise
        return df

    @staticmethod
    def tsv(path, header=True, infer_schema=True, *args, **kwargs):
        """
        Return a dataframe from a tsv file.
        :param path: path or location of the file.
        :param header: tell the function whether dataset has a header row. True default.
        :param infer_schema: infers the input schema automatically from data.
        It requires one extra pass over the data. True default.

        :return:
        """

        return Load.csv(path, sep='\t', header=header, infer_schema=infer_schema, *args, **kwargs)

    @staticmethod
    def csv(path, sep=',', header=True, infer_schema=True, na_values=None, encoding="utf-8", n_rows=-1, cache=False,
            quoting=0, lineterminator=None, error_bad_lines=False, engine="c", keep_default_na=False,
            na_filter=False, null_value=None, storage_options=None, conn=None, n_partitions=1, *args, **kwargs):

        """
        Return a dataframe from a csv file. It is the same read.csv Spark function with some predefined
        params

        :param na_filter:
        :param path: path or location of the file.
        :param sep: usually delimiter mark are ',' or ';'.
        :param header: tell the function whether dataset has a header row. True default.
        :param infer_schema: infers the input schema automatically from data.
        :param encoding:
        :param na_values:
        :param n_rows:
        :param quoting:
        :param engine: 'python' or 'c'. 'python' slower but support better error handling
        :param lineterminator:
        :param error_bad_lines:
        :param keep_default_na:
        :param cache: If calling from a url we cache save the path to the temp file so we do not need to download the file again

        """

        path = unquote_path(path)

        if cache is False:
            prepare_path.cache_clear()

        if conn is not None:
            path = conn.path(path)
            storage_options = conn.storage_options

        remove_param = "chunk_size"
        if kwargs.get(remove_param):
            # This is handle in this way to preserve compatibility with others dataframe technologies.
            logger.print(f"{remove_param} is not supported. Used to preserve compatibility with Optimus Pandas")
            kwargs.pop(remove_param)
        # if engine=="pandas":

        try:
            # From the panda docs using na_flailter
            # Detect missing value markers (empty strings and the value of na_values). In data without any NAs,
            # passing na_filter=False can improve the performance of reading a large file.
            if engine == "python":

                # na_filter=na_filter, error_bad_lines and low_memory are not support by pandas engine
                dfd = dd.read_csv(path, sep=sep, header=0 if header else None, encoding=encoding,
                                  quoting=quoting, lineterminator=lineterminator,
                                  keep_default_na=True, na_values=None, engine=engine,
                                  storage_options=storage_options, error_bad_lines=False, *args, **kwargs)

            elif engine == "c":
                dfd = dd.read_csv(path, sep=sep, header=0 if header else None, encoding=encoding,
                                  quoting=quoting, lineterminator=lineterminator, error_bad_lines=error_bad_lines,
                                  keep_default_na=True, na_values=None, engine=engine, na_filter=na_filter,
                                  storage_options=storage_options, low_memory=False, *args, **kwargs)

            if n_rows > -1:
                dfd = dd.from_pandas(dfd.head(n=n_rows), npartitions=1).reset_index(drop=True)

            df = DaskDataFrame(dfd)
            df.meta = Meta.set(df.meta, value={"file_name": path, "name": ntpath.basename(path),
                                               "max_cell_length": df.cols.len("*").cols.max()})
        except IOError as error:
            logger.print(error)
            raise

        return df

    @staticmethod
    def parquet(path, columns=None, engine="pyarrow", storage_options=None, conn=None, *args, **kwargs):
        """
        Return a dataframe from a parquet file.
        :param path: path or location of the file. Must be string dataType
        :param columns: select the columns that will be loaded. In this way you do not need to load all the dataframe
        :param engine:
        :param args: custom argument to be passed to the spark parquet function
        :param kwargs: custom keyword arguments to be passed to the spark parquet function
        :return: Spark Dataframe
        """

        path = unquote_path(path)

        if conn is not None:
            path = conn.path(path)
            storage_options = conn.storage_options

        try:
            dfd = dd.read_parquet(path, columns=columns, engine=engine, storage_options=storage_options, *args,
                                  **kwargs)
            df = DaskDataFrame(dfd)
            df.meta = Meta.set(df.meta, "file_name", path)

        except IOError as error:
            logger.print(error)
            raise

        return df

    @staticmethod
    def zip(path, sep=',', header=True, infer_schema=True, charset="UTF-8", null_value="None", n_rows=-1,
            storage_options=None, conn=None, *args, **kwargs):

        path = unquote_path(path)

        file, file_name = prepare_path(path, "zip")

        from zipfile import ZipFile
        import dask.dataframe as dd
        import os

        wd = '/path/to/zip/files'
        file_list = os.listdir(wd)
        destdir = '/extracted/destination/'

        dfd = dd.from_pandas(pd.DataFrame())

        for f in file_list:
            with ZipFile(wd + f, "r") as zip:
                zip.extractall(destdir, None, None)
                df = dd.read_csv(zip.namelist(), usecols=['Enter', 'Columns', 'Here'], parse_dates=['Date'])
                dfd = optimus.helpers.functions_spark.append(df)

        dfd.compute()

        try:
            df = dd.read_csv(file, sep=sep, header=0 if header else None, encoding=charset, na_values=null_value,
                             compression="gzip", *args, **kwargs)

            if n_rows > -1:
                df = df.rows.limit(n_rows)

            df.meta = Meta.set(df.meta, "file_name", file_name)
        except IOError as error:
            logger.print(error)
            raise
        return df

    @staticmethod
    def orc(path, columns=None, cache=None, storage_options=None, conn=None, *args, **kwargs):

        """
        Optimized Row Columnar
        Return a dataframe from a .orc file.
        params


        """

        path = unquote_path(path)

        if cache is False:
            prepare_path.cache_clear()

        if conn is not None:
            path = conn.path(path)
            storage_options = conn.storage_options

        try:
            # From the panda docs using na_filter
            # Detect missing value markers (empty strings and the value of na_values). In data without any NAs,
            # passing na_filter=False can improve the performance of reading a large file.
            dfd = dd.read_orc(path, columns, storage_options=storage_options, *args, **kwargs)

            df = DaskDataFrame(dfd)
            df.meta = Meta.set(df.meta, value={"file_name": path, "name": ntpath.basename(path)})
        except IOError as error:
            logger.print(error)
            raise

        return df

    @staticmethod
    def avro(path, storage_options=None, conn=None, *args, **kwargs):
        """
        Return a dataframe from a avro file.
        :param path: path or location of the file. Must be string dataType
        :param storage_options:
        :param args: custom argument to be passed to the avro function
        :param kwargs: custom keyword arguments to be passed to the avro function
        :return: Spark Dataframe
        """

        path = unquote_path(path)

        if conn is not None:
            path = conn.path(path)
            storage_options = conn.storage_options

        file, file_name = prepare_path(path, "avro")

        try:
            df = dask_bag.read_avro(path, storage_options=storage_options, *args, **kwargs).to_dataframe()
            df.meta = Meta.set(df.meta, "file_name", file_name)

        except IOError as error:
            logger.print(error)
            raise

        return df

    @staticmethod
    def excel(path, sheet_name=0, merge_sheets=False, skiprows=1, n_rows=-1, n_partitions=1, storage_options=None,
              conn=None, *args, **kwargs):
        """
        Return a dataframe from a excel file.
        :param path: Path or location of the file. Must be string dataType
        :param sheet_name: excel sheet name
        :param merge_sheets:
        :param args: custom argument to be passed to the excel function
        :param kwargs: custom keyword arguments to be passed to the excel function

        """

        path = unquote_path(path)

        if conn is not None:
            path = conn.path(path)
            storage_options = conn.storage_options

        file, file_name = prepare_path(path)
        header = None
        if merge_sheets is True:
            skiprows = -1
        else:
            header = 0
            skiprows = 0

        if n_rows == -1:
            n_rows = None

        pdfs = pd.read_excel(file, sheet_name=sheet_name, header=header, skiprows=skiprows, nrows=n_rows,
                             storage_options=storage_options, *args, **kwargs)
        sheet_names = list(pd.read_excel(file, None, storage_options=storage_options).keys())

        pdf = pd.concat(val_to_list(pdfs), axis=0).reset_index(drop=True)

        df = dd.from_pandas(pdf, npartitions=1)
        df.meta = Meta.set(df.meta, "file_name", ntpath.basename(file_name))
        df.meta = Meta.set(df.meta, "sheet_names", sheet_names)

        return df



# DataFrame = pd.DataFrame
from datetime import datetime, timedelta
from optimus.infer import is_list, is_list_or_tuple
from optimus.helpers.core import val_to_list

import numpy as np
import pandas as pd
import re
from optimus.engines.base.commons.functions import to_string, to_integer, to_float, to_boolean, word_tokenize
from optimus.engines.base.functions import Functions


class PandasFunctions(Functions):

    def _to_float(self, value):
        return value.map(to_float)

    def to_float(self, series):
        return to_float(series)

    def to_string(self, series):
        return to_float(series)

    def _to_integer(self, value):
        return value.map(to_integer)

    def to_integer(self, series):
        return to_integer(series)

    def to_boolean(self, series):
        return to_boolean(series)

    def to_string(self, series):
        return to_string(series)

    def word_tokenize(self, value):
        return word_tokenize(value)

    def count_zeros(self, series, *args):
        return int((self.to_float(series).values == 0).sum())

    def kurtosis(self, series):
        return self.to_float(series).kurtosis()

    def skew(self, series):
        return self.to_float(series).skew()

    def exp(self, series):
        return np.exp(self.to_float(series))

    def sqrt(self, series):
        return np.sqrt(self.to_float(series))

    def reciprocal(self, series):
        return np.reciprocal(self.to_float(series))

    def radians(self, series):
        return np.radians(self.to_float(series))

    def degrees(self, series):
        return np.degrees(self.to_float(series))

    def ln(self, series):
        return np.log(self.to_float(series))

    def log(self, series, base=10):
        return np.log(self.to_float(series)) / np.log(base)

    def sin(self, series):
        return np.sin(self.to_float(series))

    def cos(self, series):
        return np.cos(self.to_float(series))

    def tan(self, series):
        return np.tan(self.to_float(series))

    def asin(self, series):
        return np.arcsin(self.to_float(series))

    def acos(self, series):
        return np.arccos(self.to_float(series))

    def atan(self, series):
        return np.arctan(self.to_float(series))

    def sinh(self, series):
        return np.arcsinh(self.to_float(series))

    def cosh(self, series):
        return np.cosh(self.to_float(series))

    def tanh(self, series):
        return np.tanh(self.to_float(series))

    def asinh(self, series):
        return np.arcsinh(self.to_float(series))

    def acosh(self, series):
        return np.arccosh(self.to_float(series))

    def atanh(self, series):
        return np.arctanh(self.to_float(series))

    def floor(self, series):
        return np.floor(self.to_float(series))

    def ceil(self, series):
        return np.ceil(self.to_float(series))

    def replace_chars(self, series, search, replace_by):
        search = list(map(re.escape, search))
        return self.to_string(series).replace(search, replace_by, regex=True)

    def normalize_chars(self, series):
        return series.str.normalize("NFKD").str.encode('ascii', errors='ignore').str.decode('utf8')

    def date_format(self, series, current_format=None, output_format=None):
        return pd.to_datetime(series, format=current_format, errors="coerce").dt.strftime(output_format).reset_index(
            drop=True)

    def days_between(self, series, value=None, date_format=None):

        value_date_format = date_format

        if is_list_or_tuple(date_format) and len(date_format) == 2:
            date_format, value_date_format = date_format

        if is_list_or_tuple(value) and len(value) == 2:
            value, value_date_format = value

        date = pd.to_datetime(series, format=date_format, errors="coerce")
        value = pd.Timestamp.now() if value is None else pd.to_datetime(value, format=value_date_format, errors="coerce")
        
        return (value - date).dt.days



import numpy as np

from abc import abstractmethod, ABC
from optimus.helpers.types import MaskDataFrameType

from optimus.helpers.columns import parse_columns
from optimus.helpers.constants import ProfilerDataTypes
from optimus.helpers.core import val_to_list, one_list_to_val
from optimus.helpers.raiseit import RaiseIt
from optimus.infer import is_str, regex_http_code, regex_social_security_number, regex_phone_number, \
    regex_credit_card_number, regex_zip_code, regex_gender, regex_ip, regex_email, \
    is_datetime, is_list, is_bool, is_object, regex_full_url


class Mask(ABC):
    def __init__(self, root):
        self.root = root

    def _to_frame(self, series):
        if callable(getattr(series, "to_frame", False)):
            return series.to_frame()
        return series

    @abstractmethod
    def str(self, cols="*") -> MaskDataFrameType:
        pass

    @abstractmethod
    def int(self, cols="*") -> MaskDataFrameType:
        pass

    @abstractmethod
    def float(self, cols="*") -> MaskDataFrameType:
        pass

    @abstractmethod
    def numeric(self, cols="*") -> MaskDataFrameType:
        pass

    def greater_than(self, cols="*", value=None) -> MaskDataFrameType:
        df = self.root
        return df[cols] > value

    def greater_than_equal(self, cols="*", value=None) -> MaskDataFrameType:
        df = self.root
        return df[cols] >= value

    def less_than(self, cols="*", value=None) -> MaskDataFrameType:
        df = self.root
        return df[cols] < value

    def less_than_equal(self, cols="*", value=None) -> MaskDataFrameType:
        df = self.root
        return df[cols] <= value

    def equal(self, cols="*", value=None) -> MaskDataFrameType:
        df = self.root
        return df[cols] == value

    def not_equal(self, cols="*", value=None) -> MaskDataFrameType:
        df = self.root
        return df[cols] != value

    def missing(self, cols="*") -> MaskDataFrameType:
        """
        Return missing values
        :param cols:
        :return:
        """
        return self.null(cols) | self.empty(cols)

    def mismatch(self, cols="*", dtype=None) -> MaskDataFrameType:
        """
        Return missing values
        :param cols:
        :param dtype:
        :return:
        """

        df = self.root
        cols = one_list_to_val(parse_columns(df, cols))

        mask_match = df[cols].mask.match(cols, dtype)
        mask_null = df[cols].mask.null(cols)
        return ~(mask_match | mask_null)

    def match(self, cols="*", regex=None, dtype=None) -> MaskDataFrameType:
        if dtype is None:
            return self.match_regex(cols=cols, regex=regex)
        else:
            return self.match_dtype(cols=cols, dtype=dtype)

    def match_regex(self, cols="*", regex="") -> MaskDataFrameType:
        return self.root[cols].cols.apply(cols, self.root.cols.F.match, args=(regex,))

    def match_dtype(self, cols="*", dtype=None) -> MaskDataFrameType:
        """
        Return values that match with a datatype
        :param cols:
        :param dtype:
        :return:
        """
        df = self.root
        cols = one_list_to_val(parse_columns(df, cols))

        if dtype is None:
            profiled_dtype = df.profile.dtypes(cols)
            if profiled_dtype is None:
                RaiseIt.value_error(dtype, ProfilerDataTypes.list(),
                                    "dtype not found in cache. Need to passed as dtype param")
            else:
                dtype = profiled_dtype

        if is_list(dtype):
            mask_match = None
            for i, j in zip(cols, dtype):
                if mask_match is None:
                    mask_match = getattr(df[i].mask, j)(i)
                else:
                    mask_match[i] = getattr(df[i].mask, j)(i)
        else:
            mask_match = getattr(df[cols].mask, dtype)(cols)

        return mask_match

    def value_in(self, cols="*", values=None) -> MaskDataFrameType:
        
        df = self.root
        cols = parse_columns(df, cols)

        values = val_to_list(values)

        mask = df.data[cols].isin(values)
        return df.new(self._to_frame(mask))

    def pattern(self, cols="*", pattern=None) -> MaskDataFrameType:
        
        df = self.root
        cols = parse_columns(df, cols)
        
        return df[cols].cols.pattern() == pattern

    def starts_with(self, cols="*", value=None) -> MaskDataFrameType:
                
        df = self.root
        cols = val_to_list(parse_columns(df, cols))

        mask = None

        for col in cols:
            series = df.data[col].str.startswith(value, na=False)
            if mask is None:
                mask = self._to_frame(series)
            else:
                mask[col] = series


        return df.new(self._to_frame(mask))

    def ends_with(self, cols="*", value=None) -> MaskDataFrameType:
                
        df = self.root
        cols = val_to_list(parse_columns(df, cols))

        mask = None

        for col in cols:
            series = df.data[col].str.endswith(value, na=False)
            if mask is None:
                mask = self._to_frame(series)
            else:
                mask[col] = series

        return df.new(self._to_frame(mask))

    def contains(self, cols="*", value=None, case=True, flags=0, na=False, regex=False) -> MaskDataFrameType:
                
        df = self.root
        cols = val_to_list(parse_columns(df, cols))

        mask = None

        for col in cols:
            series = df.data[col].str.contains(value, case=case, flags=flags, na=na, regex=regex)
            if mask is None:
                mask = self._to_frame(series)
            else:
                mask[col] = series
        
        if is_list(value):
            value = "|".join(value)
            regex = True
        
        return df.new(self._to_frame(mask))

    def expression(self, where=None, cols="*") -> MaskDataFrameType:
        """
        Find rows and appends resulting mask to the dataset
        :param where: Mask, expression or name of the column to be taken as mask
        :param output_col:
        :return: Optimus Dataframe
        """

        df = self.root.cols.select(cols)

        if is_str(where):
            if where in df.cols.names():
                where = df[where]
            else:
                where = eval(where)

        if isinstance(where, (df.__class__,)):
            return where

        RaiseIt.type_error(where, ["Dataframe", "Expression", "Column name"])

    def find(self, cols="*", value=None) -> MaskDataFrameType:
        
        df = self.root
        cols = val_to_list(parse_columns(df, cols))

        
        if is_str(value):

            mask = None
            for col in cols:
                series = self.root.data[col].astype(str).str.match(value, na=False)
                if mask is None:
                    mask = self._to_frame(series)
                else:
                    mask[col] = series
                    
            return self.root.new(self._to_frame(mask))

        else:
            return df[cols] == value



    def null(self, cols="*", how="any") -> MaskDataFrameType:
        """
        Find the rows that have null values
        :param how:
        :param cols:
        :return:
        """
        df = self.root
        dfd = self.root.data
        cols = val_to_list(parse_columns(df, cols))
        subset_df = dfd[cols]

        if how == "all":
            col_name = cols[0] if len(cols)==1 else "__null__"
            mask = subset_df.isnull().all(axis=1).rename(col_name)
        else:
            mask = subset_df.isnull()

        return self.root.new(self._to_frame(mask))

    
    def none(self, cols="*") -> MaskDataFrameType:
        """
        Find the rows that have None values

        :param cols:
        :return:
        """
        return ~self.numeric(cols) & self.null(cols)

    def nan(self, cols="*") -> MaskDataFrameType:
        """
        Find the rows that have np.nan values

        :param cols:
        :return:
        """
        return self.numeric(cols) & self.null(cols)

    def duplicated(self, cols, keep="first") -> MaskDataFrameType:
        """
        Find the rows that have duplicated values

        :param keep:
        :param cols:
        :return:
        """
        df = self.root
        dfd = self.root.data
        cols = val_to_list(parse_columns(df, cols))

        if cols is not None:
            subset = cols
            subset_df = dfd[subset]
        else:
            subset_df = dfd

        col_name = cols[0] if len(cols)==1 else "__duplicated__"

        mask = subset_df.duplicated(keep=keep, subset=cols).rename(col_name)

        return self.root.new(self._to_frame(mask))

    def unique(self, cols, keep="first") -> MaskDataFrameType:
        """
        Find the rows that are not duplicated values

        :param keep:
        :param cols:
        :return:
        """
        return ~self.duplicated(cols, keep)

    def empty(self, cols="*") -> MaskDataFrameType:
        """
        Find the rows that do not have any info

        :param cols:
        :return:
        """
        return self.root[cols] == ""

    def email(self, cols="*") -> MaskDataFrameType:
        return self.match_regex(cols, regex_email)

    def ip(self, cols="*") -> MaskDataFrameType:
        return self.match_regex(cols, regex_ip)

    def url(self, cols="*") -> MaskDataFrameType:
        return self.match_regex(cols, regex_full_url)

    def gender(self, cols="*") -> MaskDataFrameType:
        return self.match_regex(cols, regex_gender)

    def boolean(self, cols="*") -> MaskDataFrameType:
        return self.root[cols].cols.apply(cols, is_bool)

    def zip_code(self, cols="*") -> MaskDataFrameType:
        return self.match_regex(cols, regex_zip_code)

    def credit_card_number(self, cols="*") -> MaskDataFrameType:
        return self.match_regex(cols, regex_credit_card_number)

    def datetime(self, cols="*") -> MaskDataFrameType:
        # df = self.root
        # if df[cols].cols.dtype  == df.constants.
        return self.root[cols].cols.apply(cols, is_datetime)
        # return self.match_regex(cols, regex_date)

    def object(self, cols="*") -> MaskDataFrameType:
        return self.root[cols].cols.apply(cols, is_object)
        # return self.match_regex(cols, is_object)

    def array(self, cols="*") -> MaskDataFrameType:
        return self.root[cols].cols.apply(cols, is_list)
        # return self.match_regex(cols, is_list_value)

    def phone_number(self, cols="*") -> MaskDataFrameType:
        return self.match_regex(cols, regex_phone_number)

    def social_security_number(self, cols="*") -> MaskDataFrameType:
        return self.match_regex(cols, regex_social_security_number)

    def http_code(self, cols="*") -> MaskDataFrameType:
        return self.match_regex(cols, regex_http_code)

    #

    def all(self, cols="*") -> MaskDataFrameType:
        
        df = self.root

        mask = None

        for col in df.cols.names(cols):
            _mask = df[col]
            if mask is None:
                mask = _mask
            else:
                mask = mask & _mask

        return mask

    def any(self, cols="*") -> MaskDataFrameType:
        
        df = self.root

        mask = None

        for col in df.cols.names(cols):
            _mask = df[col]
            if mask is None:
                mask = _mask
            else:
                mask = mask | _mask

        return mask



import numpy as np
from numba import njit

# Reference https://stackoverflow.com/questions/52673285/performance-of-pandas-apply-vs-np-vectorize-to-create-new-column-from-existing-c
# Reference https://stackoverflow.com/questions/12200580/numpy-function-for-simultaneous-max-and-min

@njit(fastmath=True)
def min_max(arr):
    n = arr.size
    odd = n % 2
    if not odd:
        n -= 1
    max_val = min_val = arr[0]
    i = 1
    while i < n:
        x = arr[i]
        y = arr[i + 1]
        if x > y:
            x, y = y, x
        min_val = np.nanmin([x, min_val])
        max_val = np.nanmax([y, max_val])
        i += 2
    if not odd:
        x = arr[n]
        min_val = np.nanmin([x, min_val])
        max_val = np.nanmax([y, max_val])
    return min_val, max_val


# Histogram Implementation
# Reference https://numba.pydata.org/numba-examples/examples/density_estimation/histogram/results.html
# Reference https://iscinumpy.gitlab.io/post/histogram-speeds-in-python/

# Numpy
def histogram1d(v, bins, range):
    return np.histogram(v, bins, range)


# Numba
@njit(fastmath=True)
def get_bin_edges(a, bins):
    """
    Get edges from a array
    :param a:
    :param bins:
    :return:
    """
    bin_edges = np.zeros((bins + 1,), dtype=np.float64)
    a_min, a_max = min_max(a)
    delta = (a_max - a_min) / bins
    for i in range(bin_edges.shape[0]):
        bin_edges[i] = a_min + i * delta

    bin_edges[-1] = a_max  # Avoid round off error on last point
    return bin_edges


@njit(fastmath=True)
def get_bin_edges_min_max(a_min, a_max, bins):
    bin_edges = np.zeros((bins + 1,), dtype=np.float64)
    # a_min, a_max = min , max
    # a_min = a.min()
    # a_max = a.max()
    delta = (a_max - a_min) / bins
    for i in range(bin_edges.shape[0]):
        bin_edges[i] = a_min + i * delta

    bin_edges[-1] = a_max  # Avoid roundoff error on last point
    return bin_edges


@njit(fastmath=True)
def compute_bin(x, bin_edges):
    # assuming uniform bins for now
    n = bin_edges.shape[0] - 1
    a_min = bin_edges[0]
    a_max = bin_edges[-1]

    # special case to mirror NumPy behavior for last bin
    if x == a_max:
        return n - 1  # a_max always in last bin

    bin = int(n * (x - a_min) / (a_max - a_min))

    if bin < 0 or bin >= n:
        return None
    else:
        return bin


# Dask. This NOT work faster than histogram1 library. Require more testing
# numba_histogram1 = delayed(numba_histogram)
# # print(func(df["OFFENSE_CODE"],10).compute())
# %timeit numba_histogram1(np.array(df["OFFENSE_CODE"]),10).compute()

@njit(fastmath=True)
def numba_histogram(a, bins):

    hist = np.zeros((bins,), dtype=np.intp)
    bin_edges = get_bin_edges(a, bins)

    for x in a.flat:
        bin = compute_bin(x, bin_edges)
        if bin is not None:
            hist[int(bin)] += 1

    return hist, bin_edges

@njit
def numba_count_uniques(a):
    q = np.zeros(len(a), dtype=np.uint8)
    count = 0
    for x in a.flat:
        if q[x] == 0:
            q[x] = 1
            count += 1
    return count

@njit(fastmath=True)
def numba_histogram_edges(a, bin_edges):
    hist = np.zeros(len(bin_edges), dtype=np.intp)

    for x in a.flat:
        bin = compute_bin(x, bin_edges)
        if bin is not None:
            hist[int(bin)] += 1

    return list(hist)


def count_na_j(df, series):
    return np.count_nonzero(df[series].isnull().values.ravel())


# Reference https://stackoverflow.com/questions/10741346/numpy-most-efficient-frequency-counts-for-unique-values-in-an-array
# https://stackoverflow.com/a/21124789/581858
def bincount(a, n):
    y = np.bincount(a)
    ii = np.nonzero(y)[0]
    r = np.vstack((ii, y[ii])).T
    r = r[r[:, 1].argsort()[::-1]][:n]
    i, j = np.hsplit(r, 2)
    i, j = np.concatenate(i), np.concatenate(j)

    return i, j

    # Old implementation
    # i, j = np.unique(df[col_name], return_counts=True)
    # count_sort_ind = np.argsort(-j)
